const deepEqual = require('deep-equal')
const stringify = require('json-stable-stringify')

// properties - the properties that correspond to types
// types - the expected types of the properties
// returns list of properties found matching order of types

const args = (context, hierarchy) => ({ types, properties }) => {
  const orderedByType = []

  for (const type of types) {
    let found = false
    for (let index = 0; index < properties.length; ++index) {
      const value = context[properties[index]]
      if (!value) {
        return false
      }
      if (hierarchy.isA(value.marker, type)) {
        orderedByType.push(properties[index])
        properties.splice(index, 1)
        found = true
        break
      }
    }
    if (!found) {
      return
    }
  }

  if (orderedByType.length === 0) {
    return
  }

  return orderedByType
}

const appendNoDups = (l1, l2) => {
  for (const x of l2) {
    if (l1.includes(x)) {
      continue
    }
    l1.push(x)
  }
}

const safeEquals = (v1, v2) => {
  if (!deepEqual(stringify(v1), stringify(v2))) {
    return false
  }
  return true
}

/*
const semanticsGenerate = (from, known) => {
  const marker = from.marker

  const sources = []
  for (let key in Object.keys(from)) {
    if (from[key].marker) {
      sources.push(from[key])
    }
  }
  const mappings = []
  for (let key in Object.keys(to)) {
    const source
    if (to[keys].marker ==
  }

  return {
    match: ({context}) => marker == marker,
    apply: ({context}) => {
    },
  }
}
*/

const hashIndexesGet = (hash, indexes) => {
  let value = hash
  for (const i of indexes) {
    value = value[i]
  }
  return value
}

const hashIndexesSet = (hash, indexes, value) => {
  let currentValue = hash
  for (const i of indexes.slice(0, -1)) {
    if (!currentValue[i]) {
      currentValue[i] = {}
    }
    currentValue = currentValue[i]
  }
  currentValue[indexes[indexes.length - 1]] = value
}

const translationMapping = (from, to) => {
  const mappings = []
  for (const fkey of Object.keys(from)) {
    if (from[fkey].value) {
      let found = false
      const todo = Object.keys(to).map((key) => [key])
      while (!found) {
        const tkey = todo.shift()
        const tvalue = hashIndexesGet(to, tkey)
        const fvalue = hashIndexesGet(from, [fkey])
        if (fvalue.value === tvalue.value) {
          mappings.push({ from: [fkey], to: tkey })
          found = true
          break
        } else {
          if (typeof tvalue !== 'string' && typeof tvalue !== 'number') {
            for (const key of Object.keys(tvalue)) {
              todo.push(tkey.concat(key))
            }
          }
        }
      }
    }
  }
  return mappings
}

const normalizeGenerator = (generator) => {
  if (Array.isArray(generator)) {
    if (generator.length === 2) {
      return { match: generator[0], apply: generator[1] }
    } else {
      return { match: generator[0], apply: generator[1], uuid: generator[2] }
    }
  }
  return generator
}

const normalizeSemantic = (semantic) => {
  if (Array.isArray(semantic)) {
    if (semantic.length === 2) {
      return { match: semantic[0], apply: semantic[1] }
    } else {
      return { match: semantic[0], apply: semantic[1], uuid: semantic[2] }
    }
  }
  return semantic
}

const isArray = (value) => {
  return (!!value) && (value.constructor === Array)
}

const isObject = (value) => {
  return (!!value) && (value.constructor === Object)
}

const isCompound = (value) => {
  return isArray(value) || isObject(value)
}

nextCallId = 0
nextContextId = 0

class InitCalls {

  constructor() {
    this.nextCallId = 0
    this.nextContextId = 0
    this.stack = []
  }

  start() {
    return this.nextCallId
  }

  next() {
    // this.nextContextId += 1
    this.nextContextId += 1
  }

  push() {
    this.nextCallId += 1
    // this.nextCallId += 1
    // this.stack.push(this.nextCallId)
    this.stack.push(this.nextCallId)
    const calls = this.stack.map( (call) => `call${call}` )
    // return `Context#${this.nextContextId}: ${calls}`
    return `Context#${nextContextId}: ${calls}`
  }

  current() {
    return `call${this.stack[this.stack.length-1]}`
  }

  touch(context) {
    if (!context.touchedBy) {
      context.touchedBy = []
    }
    if (!context.touchedBy.includes( this.current() )) {
      context.touchedBy.push( this.current() )
    }
  }

  pop() {
    this.stack.pop()
  }
}

const hashCode = (str) => {
  var hash = 0, i, ch;
  if (str.length === 0) return hash;
  for (i = 0; i < str.length; i++) {
    ch   = str.charCodeAt(i);
    hash  = ((hash << 5) - hash) + ch;
    hash |= 0; // Convert to 32bit integer
  }
  return hash;
};

const sortJson = (json) => {
  return json
}

const validProps = (valids, object, type) => {
  for (let prop of Object.keys(object)) {
    let okay = false
    for (valid of valids) {
      if (prop.match(valid)) {
        okay = true
        break
      }
    }
    if (!okay) {
      throw `Unknown property "${prop}" in the ${type}. Valid properties are ${valid}. The ${type} is ${JSON.stringify(object)}`
    }
  }
}

const mapInPlace = (list, fn) => {
  for (let i = 0; i < list.length; ++i) {
    list[i] =fn(list[i])
  }
}

module.exports = { 
  mapInPlace,
  validProps, 
  args, 
  safeEquals, 
  appendNoDups, 
  hashIndexesGet, 
  hashIndexesSet, 
  translationMapping, 
  normalizeGenerator, 
  normalizeSemantic, 
  isArray, 
  isObject, 
  isCompound, 
  InitCalls, 
  hashCode, 
  sortJson 
}
